/* MinuteMonitor, log duration and display usage for the N900
 * Copyright (c) 2012 David Fries <David@Fries.net> http://david.fries.net/
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "DBMonitor.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <sstream>
#include <sqlite3.h>

using namespace std;

DBMonitor::DBMonitor() :
	Verbose(0),
	DbCall(NULL),
	DbConfig(NULL),
	CallStmt(NULL),
	timestamp(time(NULL)),
	minutes_available(0),
	period_end(1),
	evening_end(0),
	evening_begin(24*60*60-1),
	incoming_free(false),
	nonprime_free(true)
{
}

DBMonitor::~DBMonitor()
{
	Close();
}

bool DBMonitor::Open()
{
	const char *home=getenv("HOME");
	if(!home)
		return false;

	string file=string(home)+"/.rtcom-eventlogger/el-v1.db";
	if(sqlite3_open_v2(file.c_str(), &DbCall, SQLITE_OPEN_READONLY, NULL))
	{
		cerr << "Error opening " << file << ' '
			<< sqlite3_errmsg(DbCall) << endl;
		cerr << "$HOME currently set to " << home << endl;
		return false;
	}
	// The two ?'s are bound later to the specific time interval.
	if(sqlite3_prepare_v2(DbCall, "select start_time, end_time, "
		"Events.local_uid, Events.remote_uid, Remotes.remote_name, "
		"outgoing "
		"from Events left outer join Remotes "
		"using (local_uid, remote_uid), Plugins "
		"where service_id = Plugins.id and Plugins.name='CALL' "
		" and start_time >= ? and start_time <= ? order by Events.id;",
		-1, &CallStmt, NULL))
	{
		cerr << "Error preparing the statement "
			<< sqlite3_errmsg(DbConfig) << endl;
		return false;
	}

	file=string(home)+"/.rtcom-eventlogger/MinuteMonitor.db";
	if(sqlite3_open_v2(file.c_str(), &DbConfig,
		SQLITE_OPEN_READWRITE, NULL))
	{
		if(sqlite3_open_v2(file.c_str(), &DbConfig,
			SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, NULL))
		{
			cerr << "Error opening " << file << ' '
				<< sqlite3_errmsg(DbConfig) << endl;
			return false;
		}
		else
		{
			if(!CreateConfigTables())
				return false;
		}
	}
	// Give it two chances, if the first try fails it will create or
	// update the tables which can allow the second try to succeed.
	if(!LoadSettings())
		LoadSettings();
	LoadFreeContacts();

	return true;
}

bool DBMonitor::CreateConfigTables()
{
	// This routine is called when a new database is created or when
	// something failed, like preparing a statement.  If the Settings
	// table is out of date and needs additional columns most of these
	// commands will fail, but as long as "something_worked" it is
	// probably ready to try the operation again.
	bool something_worked=false;
	const char *query[]=
	{
		"ALTER TABLE Settings "
		"ADD COLUMN IncomingFree INTEGER DEFAULT 0 NOT NULL;",

		"ALTER TABLE Settings "
		"ADD COLUMN NonprimeFree INTEGER DEFAULT 1 NOT NULL;",

		"CREATE TABLE Settings ("
		"timestamp INTEGER PRIMARY KEY, "
		"minutes_available INTEGER NOT NULL, "
		"period_end INTEGER NOT NULL, "
		"evening_end INTEGER NOT NULL, "
		"evening_begin INTEGER NOT NULL, "
		"IncomingFree INTEGER NOT NULL, "
		"NonprimeFree INTEGER NOT NULL);",

		"CREATE TABLE NoChargeContacts ("
		"remote_uid TEXT PRIMARY KEY NOT NULL); ",

		"CREATE INDEX idx_nccontact_uid on "
		"NoChargeContacts(remote_uid);",
	};
	for(size_t i=0; i<sizeof(query)/sizeof(*query); ++i)
	{
		char *errmsg=NULL;
		sqlite3_exec(DbConfig, query[i], NULL, NULL, &errmsg);
		if(errmsg)
		{
			cerr << "Error creating/modifying config tables table "
				<< errmsg << endl;
		}
		else
		{
			something_worked=true;
		}
	}
	if(something_worked)
		cout << "config tables created/updated\n";
	else
		cout << __func__ << " failed no tables modified\n";
	return something_worked;
}

bool DBMonitor::LoadSettings()
{
	sqlite3_stmt *stmt=NULL;
	// get the greatest timestamp entry in the config file
	if(sqlite3_prepare_v2(DbConfig, "select timestamp, minutes_available, "
		"period_end, evening_end, evening_begin, "
		"IncomingFree, NonprimeFree from Settings "
		"order by timestamp desc limit 1;", -1, &stmt, NULL))
	{
		cerr << "Error preparing the statement "
			<< sqlite3_errmsg(DbConfig) << endl;
		CreateConfigTables();
		return false;
	}
	int result=SQLITE_ERROR;
	int rows=0;
	while((result=sqlite3_step(stmt)) == SQLITE_ROW)
	{
		if(sqlite3_column_count(stmt) == 5)
		{
			// missing IncomingFree, NonprimeFree, ...
			CreateConfigTables();
			break;
		}
		if(sqlite3_column_type(stmt, 0) != SQLITE_INTEGER ||
			sqlite3_column_type(stmt, 1) != SQLITE_INTEGER ||
			sqlite3_column_type(stmt, 2) != SQLITE_INTEGER ||
			sqlite3_column_type(stmt, 3) != SQLITE_INTEGER ||
			sqlite3_column_type(stmt, 4) != SQLITE_INTEGER ||
			sqlite3_column_type(stmt, 5) != SQLITE_INTEGER ||
			sqlite3_column_type(stmt, 6) != SQLITE_INTEGER)
		{
			cerr << "Unexpected table schema\n";
			break;
		}
		timestamp=sqlite3_column_int(stmt, 0);
		minutes_available=sqlite3_column_int(stmt, 1);
		period_end=sqlite3_column_int(stmt, 2);
		evening_end=sqlite3_column_int(stmt, 3);
		evening_begin=sqlite3_column_int(stmt, 4);
		incoming_free=sqlite3_column_int(stmt, 5);
		nonprime_free=sqlite3_column_int(stmt, 6);
		++rows;
	}

	sqlite3_finalize(stmt);

	// Assume the tables need to be created.
	if(result==SQLITE_ERROR)
		CreateConfigTables();
	return rows;
}

const DBMonitor::LogType& DBMonitor::GetAllTelContacts()
{
	if(AllTelContacts.empty())
		LoadAllTelContacts();
	return AllTelContacts;
}

bool DBMonitor::LoadAllTelContacts()
{
	sqlite3_stmt *stmt=NULL;
	if(sqlite3_prepare_v2(DbCall, "select remote_uid, remote_name "
		"from Remotes where local_uid='ring/tel/ring';",
		-1, &stmt, NULL))
	{
		cerr << "Error preparing the statement "
			<< sqlite3_errmsg(DbCall) << endl;
		return false;
	}
	int result=SQLITE_ERROR;
	AllTelContacts.clear();
	LogEntry entry={0, 0, 0, false, CALL_NEITHER, "", "", UNKNOWN_RATE};
	const char *remote_name;
	while((result=sqlite3_step(stmt)) == SQLITE_ROW)
	{
		if(sqlite3_column_type(stmt, 0) != SQLITE_TEXT ||
			(sqlite3_column_type(stmt, 1) != SQLITE_TEXT &&
			sqlite3_column_type(stmt, 1) != SQLITE_NULL))
		{
			cerr << "Unexpected table schema\n";
			break;
		}
		entry.remote_uid=(const char*)sqlite3_column_text(stmt, 0);
		remote_name=(const char*)sqlite3_column_text(stmt, 1);
		entry.remote_name=remote_name ? remote_name : "";
		AllTelContacts.push_back(entry);
	}

	sqlite3_finalize(stmt);

	return result==SQLITE_DONE;
}

bool DBMonitor::LoadFreeContacts()
{
	sqlite3_stmt *stmt=NULL;
	if(sqlite3_prepare_v2(DbConfig,
		"select remote_uid from NoChargeContacts;", -1, &stmt, NULL))
	{
		cerr << "Error preparing the statement "
			<< sqlite3_errmsg(DbConfig) << endl;
		return false;
	}
	int result=SQLITE_ERROR;
	FreeContacts.clear();
	while((result=sqlite3_step(stmt)) == SQLITE_ROW)
	{
		if(sqlite3_column_type(stmt, 0) != SQLITE_TEXT)
		{
			cerr << "Unexpected table schema\n";
			break;
		}
		FreeContacts.insert((const char*)sqlite3_column_text(stmt, 0));
	}

	sqlite3_finalize(stmt);

	return result==SQLITE_DONE;
}

bool DBMonitor::WriteConfig()
{
	int timestamp=(int)time(NULL);
	char buf[300]={0};
	snprintf(buf, sizeof(buf)-1, "insert into Settings (timestamp, "
		"minutes_available, period_end, evening_end, evening_begin, "
		"IncomingFree, NonprimeFree) "
		"values ('%d', '%d', '%d', '%d', '%d', '%d', '%d');",
		timestamp, minutes_available, period_end, evening_end,
		evening_begin, incoming_free, nonprime_free);
	char *errmsg=NULL;
	sqlite3_exec(DbConfig, buf, NULL, NULL, &errmsg);
	if(errmsg)
	{
		cerr << "Error adding to Settings table " << errmsg << endl;
		return false;
	}
	return true;
}

void DBMonitor::AddFreeContact(const std::string &remote_uid)
{
	FreeContacts.insert(remote_uid);
	char buf[200]={0};
	snprintf(buf, sizeof(buf)-1,
		"insert into NoChargeContacts (remote_uid) "
		" values ('%s');", remote_uid.c_str());
	char *errmsg=NULL;
	sqlite3_exec(DbConfig, buf, NULL, NULL, &errmsg);
	if(errmsg)
	{
		cerr << "Error inserting into NoChargeContacts table "
			<< errmsg << endl;
	}
}

void DBMonitor::RemoveFreeContact(const std::string &remote_uid)
{
	FreeContacts.erase(remote_uid);
	char buf[200]={0};
	snprintf(buf, sizeof(buf)-1,
		"delete from NoChargeContacts "
		" where remote_uid='%s';", remote_uid.c_str());
	char *errmsg=NULL;
	sqlite3_exec(DbConfig, buf, NULL, NULL, &errmsg);
	if(errmsg)
	{
		cerr << "Error deleting from NoChargeContacts table "
			<< errmsg << endl;
	}
}

void DBMonitor::Close()
{
	sqlite3_close(DbCall);
	DbCall=NULL;
	sqlite3_close(DbConfig);
	DbConfig=NULL;
	sqlite3_finalize(CallStmt);
	CallStmt=NULL;
}

bool DBMonitor::Load(time_t start, time_t end)
{
	if(!CallStmt)
		return false;
	sqlite3_reset(CallStmt);
	Log.clear();

	// start of the call, >=
	sqlite3_bind_int(CallStmt, 1, start);
	// <=
	sqlite3_bind_int(CallStmt, 2, end);

	LogEntry entry;
	entry.rate=UNKNOWN_RATE;
	const char *local_uid, *remote_name;
	// expecting results in the form, name type column,
	// start_time int 0, end_time int 1, local_uid text 2,
	// remote_uid text 3, remote_name text or NULL 4, outgoing int 5
	while(sqlite3_step(CallStmt) == SQLITE_ROW)
	{
		if(sqlite3_column_type(CallStmt, 0) != SQLITE_INTEGER ||
			sqlite3_column_type(CallStmt, 1) != SQLITE_INTEGER ||
			sqlite3_column_type(CallStmt, 2) != SQLITE_TEXT ||
			sqlite3_column_type(CallStmt, 3) != SQLITE_TEXT ||
			(sqlite3_column_type(CallStmt, 4) != SQLITE_TEXT &&
			sqlite3_column_type(CallStmt, 4) != SQLITE_NULL) ||
			sqlite3_column_type(CallStmt, 5) != SQLITE_INTEGER)
		{
			cerr << "Unexpected table schema\n";
			return false;
		}
		entry.start_time = sqlite3_column_int(CallStmt, 0);
		entry.end_time = sqlite3_column_int(CallStmt, 1);
		entry.ceil_min = ((entry.end_time - entry.start_time)+59)/60;
		if(entry.ceil_min < 0)
			entry.ceil_min = 0;
		local_uid = (const char*)sqlite3_column_text(CallStmt, 2);
		entry.remote_uid = (const char*)sqlite3_column_text(CallStmt,
			3);
		remote_name = (const char*)sqlite3_column_text(CallStmt, 4);
		// column is outgoing
		entry.direction = sqlite3_column_int(CallStmt, 5) ? CALL_OUT :
			CALL_IN;
		entry.remote_name = remote_name ? remote_name : "";
		// I assume this is the only way to identify VoIP vs not
		if(!strcmp(local_uid, "ring/tel/ring"))
			entry.VoIP=false;
		else
			entry.VoIP=true;
		Log.push_back(entry);
	}
	return true;
}
