/* MinuteMonitor, log duration and display usage for the N900
 * Copyright (c) 2012 David Fries <David@Fries.net> http://david.fries.net/
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __DB_MONITOR_H
#define __DB_MONITOR_H 1

#include <sys/time.h>
#include <stdint.h>
#include <string>
#include <map>
#include <tr1/unordered_set>
#include <set>
#include <vector>

typedef struct sqlite3 sqlite3;
typedef struct sqlite3_stmt sqlite3_stmt;

// update the call log database end_time field
class DBMonitor
{
public:
	DBMonitor();
	~DBMonitor();
	// opens the database
	bool Open();
	void Close();

	// load data, >= start and <= stop
	// the stop at the end of one time period is one second before the
	// start of the next time period, so both sides are inclusive
	bool Load(time_t start, time_t end);

	// bit field, indicate which types of calls were received (for
	// contact list which could include both)
	enum CallRateEnum {UNKNOWN_RATE=0, DAYTIME_RATE=1, EVENING_RATE=2,
		BOTH_RATE=DAYTIME_RATE | EVENING_RATE};
	// bit field, indicates direction, call lists can have calls in both
	// direction
	enum CallDirection {CALL_NEITHER=0, CALL_IN=1, CALL_OUT=2,
		CALL_BOTH= CALL_IN | CALL_OUT};
	struct LogEntry
	{
		int start_time, end_time;
		// ((end_time-start_time)+59)/60 -- ie minutes rounded up
		int ceil_min;
		bool VoIP;
		CallDirection direction;
		std::string remote_uid, remote_name; 
		// not stored in the database will always be UNKNOWN_RATE until
		// updated
		CallRateEnum rate;
	};

	typedef std::vector<LogEntry> LogType;
	const LogType& DataSet() const { return Log; }
	// allows modifying the log data
	LogType& DataSetEdit() { return Log; }
	// All telephone contacts (remote_uid), loaded the first time requested.
	const LogType& GetAllTelContacts();

	// The set of telephone contacts (remote_uid) that are classified as
	// no charge to talk to.
	typedef std::tr1::unordered_set<std::string> ContactsHashType;
	const ContactsHashType& GetFreeContacts() const { return FreeContacts; }
	bool IsFreeContact(const std::string &remote_uid) const
	{ return FreeContacts.find(remote_uid) != FreeContacts.end(); }
	// add or remove from the list and update the database
	void AddFreeContact(const std::string &remote_uid);
	void RemoveFreeContact(const std::string &remote_uid);

	void SetVerbose(int v) { Verbose=v; }
	int GetVerbose() const { return Verbose; }

	void SetIncomingFree(bool free) { incoming_free=free; }
	bool GetIncomingFree() const { return incoming_free; }
	// intended to cover evenings or weekends,
	// putting GetEveningsOrWeekendsFree gets a bit long
	void SetNonprimeFree(bool free) { nonprime_free=free; }
	bool GetNonprimeFree() const { return nonprime_free; }
	// in minutes
	void SetMinutesAvailable(int min) { minutes_available=min; }
	int GetMinutesAvailable() const { return minutes_available; }
	// day of the month
	void SetPeriodEnd(int day_of_month) { period_end=day_of_month; }
	int GetPeriodEnd() const { return period_end; }
	// seconds from the start of the day
	void SetEveningEnd(int sec) { evening_end=sec; }
	int GetEveningEnd() const { return evening_end; }
	void SetEveningBegin(int sec) { evening_begin=sec; }
	int GetEveningBegin() const { return evening_begin; }
	// Write the current configuration values to the config database.
	bool WriteConfig();

private:
	int Verbose;
	// One read only database for the log of calls received, and other
	// read write for configuration information.
	sqlite3 *DbCall, *DbConfig;
	sqlite3_stmt *CallStmt;

	// from the Settings table
	// timestamp: currently unused, intended to allow possible future
	// tracking of the plan time changes
	// minutes_available: number of minutes available per period
	// period_end: day of the month that the plan ends, which is taken
	// to be at the end of that day
	// evening_end: time in seconds during the week that end evening rate
	// ends, which will be a time in the morning
	// evening_begin: same only when it begins
	int timestamp, minutes_available, period_end,
		evening_end, evening_begin;
	// try to catch more plan types
	// incoming_free, a common option is incoming calls are free
	// nonprime_free, if evening or weekend calls are free
	bool incoming_free, nonprime_free;

	LogType Log;
	ContactsHashType FreeContacts;
	LogType AllTelContacts;

	bool CreateConfigTables();
	bool LoadSettings();
	bool LoadAllTelContacts();
	bool LoadFreeContacts();
};

#endif // __DB_MONITOR_H
