/* MinuteMonitor, log duration and display usage for the N900
 * Copyright (c) 2010-2012 David Fries <David@Fries.net> http://david.fries.net/
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "InputDBus.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <iostream>
#include <sstream>

using namespace std;

InputDBus::InputDBus(DBLog *log) :
	Log(log)
{
	// init error structure
	dbus_error_init(&Err);

	// request device adds and removals
	const char *rule[2][5]=
	{
		// system bus
		{

		// phone call, two strings, first ringing, active, none
		/*
		"type='signal',path='/com/nokia/mce/signal',"
		"interface='com.nokia.mce.signal',member='sig_call_state_ind'",
		*/

		// for testing on the desktop, just because it fires regularly
		"type='signal',path='/org/freedesktop/UPower',"
		"interface='org.freedesktop.UPower',member='DeviceChanged'",

		NULL
		},

		// session bus
		{

		// setting up call
		// object path, string, int, int, boolean
		// the object path matches the string for the call_ended
		"type='signal',"
		// POTs and SIP path starts with
		// /org/freedesktop/Telepathy/Connection
		// POTs ring/tel/ring
		// SIP sofiasip/sip/<contact>
		// so require a path_namespace partial match but
		// path_namespace requires dbus 1.5.0, N900 has 1.2.14
		// filter on receive then
		"interface='org.freedesktop.Telepathy.Connection',"
		"member=NewChannel",

		// log entry notification
		// int, string, string, string, string, string
		// int is db log entry id
		// string
		// string is contact identifier
		"type='signal',path='/rtcomeventlogger/signal',"
		"interface='rtcomeventlogger.signal',member='NewEvent'",

		// call_ended
		// string, int
		// string matches setting up call object path
		"type='signal',"
		"path='/com/nokia/policy/telephony',"
		"interface='com.nokia.policy.telephony',member='call_ended'",

		NULL
		},
	};

	// connect to the bus
	for(size_t i=0; i<sizeof(Conn)/sizeof(*Conn); ++i)
	{
		Conn[i] = dbus_bus_get(i?DBUS_BUS_SESSION:DBUS_BUS_SYSTEM,
			&Err);
		IsError("bus_get error");
		if(!Conn[i])
		{
			fprintf(stderr, "Could not open DBus\n");
			exit(1);
		}
		if(!dbus_connection_set_watch_functions(Conn[i], AddWatch,
			RemoveWatch, ToggleWatch, this, NULL))
		{
			fprintf(stderr, "Error setting up watch functions\n");
			exit(1);
		}
		// Is this needed for receiving simple signals?
		// if(!dbus_connection_set_timeout_functions(Conn, 

		// I'm using this to receive a signal, at least I'm not seeing
		// a different way to do so.
		if(!dbus_connection_add_filter(Conn[i], HandleMessage, this,
			NULL))
		{
			fprintf(stderr, "Error adding signal handler\n");
			exit(1);
		}

		for(size_t r=0; rule[i][r]; ++r)
		{
			dbus_bus_add_match(Conn[i], rule[i][r], &Err);
			dbus_connection_flush(Conn[i]);
			if(IsError("Add signal request error"))
				exit(1);
		}
		while(dbus_connection_dispatch(Conn[i])==
			DBUS_DISPATCH_DATA_REMAINS)
		{}
	}
	if(Log && Log->GetVerbose())
		printf("DBus connected\n");
}

InputDBus::~InputDBus()
{
	for(size_t i=0; i<sizeof(Conn)/sizeof(*Conn); ++i)
	{
		if(!Conn[i])
			continue;
		dbus_connection_set_watch_functions(Conn[i], NULL, NULL, NULL,
			this, NULL);
	}

	while(!WatchSet.empty())
		RemoveWatch(*WatchSet.begin(), this);
		
	for(size_t i=0; i<sizeof(Conn)/sizeof(*Conn); ++i)
	{
		if(!Conn[i])
			continue;
		dbus_connection_remove_filter(Conn[i], HandleMessage, this);
		while(dbus_connection_dispatch(Conn[i])==
			DBUS_DISPATCH_DATA_REMAINS)
		{}
		dbus_connection_unref(Conn[i]);
	}
	delete Log;
}

int InputDBus::ProcessEvents()
{
	size_t i;
	for(;;)
	{
		// -1 error, 0 timeout (not used), positive number of events
		if(poll(&PollSet[0], PollSet.size(), -1)==-1)
		{
			perror("poll error");
			break;
		}
		for(i=0; i<PollSet.size(); ++i)
		{
			pollfd &fds=PollSet[i];
			DBusWatch *watch=WatchSet[i];
			if(fds.revents & POLLIN)
				dbus_watch_handle(watch, DBUS_WATCH_READABLE);
			if(fds.revents & POLLOUT)
				dbus_watch_handle(watch, DBUS_WATCH_WRITABLE);
		}
		for(i=0; i<sizeof(Conn)/sizeof(*Conn); ++i)
		{
			while(dbus_connection_dispatch(Conn[i])==
				DBUS_DISPATCH_DATA_REMAINS)
			{}
		}
	}
	return -1;
}

DBusHandlerResult InputDBus::HandleMessage(DBusConnection * /*conn*/,
	DBusMessage *msg)
{
	const char *path=dbus_message_get_path(msg);
	const char *interface=dbus_message_get_interface(msg);
	const char *member=dbus_message_get_member(msg);
	if(Log && Log->GetVerbose())
		printf("message, path %s interface %s member %s\n",
			path, interface, member);
	int32_t i, ui1, ui2;
	dbus_bool_t b;
	const char *str1=NULL, *str2=NULL, *str3=NULL, *str4=NULL, *str5=NULL;
	const char path_connect[]="/org/freedesktop/Telepathy/Connection/";
	// check for the setting up call message
	if(!strncmp(path, path_connect, sizeof(path_connect)-1) &&
		!strcmp(interface, "org.freedesktop.Telepathy.Connection") &&
		!strcmp(member, "NewChannel") &&
		dbus_message_get_args(msg, &Err,
		DBUS_TYPE_OBJECT_PATH, &str1,
		DBUS_TYPE_STRING, &str2,
		DBUS_TYPE_UINT32, &ui1,
		DBUS_TYPE_UINT32, &ui2,
		DBUS_TYPE_BOOLEAN, &b,
		DBUS_TYPE_INVALID) && !IsError("arg decode") && str1)
	{
		Log->CallSetup(path, str1, str2, ui1, ui2, b);
	}
	// check for the setting up call message
	else if(!strcmp(path, "/rtcomeventlogger/signal") &&
		!strcmp(interface, "rtcomeventlogger.signal") &&
		!strcmp(member, "NewEvent") &&
		dbus_message_get_args(msg, &Err,
		DBUS_TYPE_INT32, &i,
		DBUS_TYPE_STRING, &str1,
		DBUS_TYPE_STRING, &str2,
		DBUS_TYPE_STRING, &str3,
		DBUS_TYPE_STRING, &str4,
		DBUS_TYPE_STRING, &str5,
		DBUS_TYPE_INVALID) && !IsError("arg decode") &&
		!IsError("arg decode") && str1 && str2 && str3 && str4 && str5)
	{
		Log->CallLog(i, str1, str2, str3, str4, str5);
	}
	// check for call_ended
	else if(!strcmp(path, "/com/nokia/policy/telephony") &&
		!strcmp(interface, "com.nokia.policy.telephony") &&
		!strcmp(member, "call_ended") &&
		dbus_message_get_args(msg, &Err,
		DBUS_TYPE_STRING, &str1,
		DBUS_TYPE_INT32, &i,
		DBUS_TYPE_INVALID) && !IsError("arg decode") && str1)
	{
		Log->CallEnded(str1, i);
	}
	// As this is a filter, don't remove the message from processing.
	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

dbus_bool_t InputDBus::AddWatch(DBusWatch *watch)
{
	// dbus_watch_get_unix_fd is documented to be preferred over *get_socket
	int fd=dbus_watch_get_unix_fd(watch);
	unsigned int flags=dbus_watch_get_flags(watch);
	int enable=dbus_watch_get_enabled(watch);
	if(Log && Log->GetVerbose())
		printf("%s watch %p fd %d %s %s\n", __func__, watch, fd,
			flags==DBUS_WATCH_READABLE?"read":"write",
			enable?"enable":"disable");
	pollfd fds;
	memset(&fds, 0, sizeof(fds));
	fds.fd=fd;
	if(enable && flags & DBUS_WATCH_READABLE)
		fds.events|=POLLIN;
	if(enable && flags & DBUS_WATCH_WRITABLE)
		fds.events|=POLLOUT;

	PollSet.push_back(fds);
	WatchSet.push_back(watch);
	return true;
}

void InputDBus::RemoveWatch(DBusWatch *watch)
{
	// dbus_watch_get_unix_fd is documented to be preferred over *get_socket
	int fd=dbus_watch_get_unix_fd(watch);
	unsigned int flags=dbus_watch_get_flags(watch);
	if(Log && Log->GetVerbose())
		printf("%s %p fd %d %s\n", __func__, watch, fd,
			flags==DBUS_WATCH_READABLE?"read":"write");
	for(size_t i=0; i<WatchSet.size(); ++i)
	{
		if(WatchSet[i]==watch)
		{
			PollSet.erase(PollSet.begin()+i);
			WatchSet.erase(WatchSet.begin()+i);
			return;
		}
	}
	fprintf(stderr, "%d not found\n", fd);
}

void InputDBus::ToggleWatch(DBusWatch *watch)
{
	int fd=dbus_watch_get_unix_fd(watch);
	//unsigned int flags=dbus_watch_get_flags(watch);
	int enable=dbus_watch_get_enabled(watch);
	if(Log && Log->GetVerbose())
		printf("%s fd %d %s\n", __func__, fd,
			enable?"enable":"disable");
	if(enable)
		AddWatch(watch);
	else
		RemoveWatch(watch);
}

bool InputDBus::IsError(const std::string &msg)
{
	if(dbus_error_is_set(&Err))
	{
		fprintf(stderr, "%s (%s)\n", msg.c_str(), Err.message);
		dbus_error_free(&Err);
		return true;
	}
	return false;
}
