/* MinuteMonitor, log duration and display usage for the N900
 * Copyright (c) 2012 David Fries <David@Fries.net> http://david.fries.net/
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <fstream>
#include <string>
#include <set>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <sqlite3.h>

using namespace std;

class Loader
{
public:
	Loader() : Db(NULL) {}
	~Loader();
	bool OpenDB(const char *file);
	// parse the given html file
	bool Parse(const char *file);
	// load entires from the database, only call after all html
	// pages have been loaded to limit the range loaded
	bool LoadDB();
	bool Match();
	void PrintMatches();
	void CompareDuration();
	bool Update();
private:
	// try to match with the given second leniency
	bool Match(int sec);
	// check for digits
	static bool Digits(const std::string &name);
	static string RemoveDash(const std::string &in);
	// normalize the string removing any leading +1 or 1
	static string Normalize(string in);
	static bool UidMatch(const string &bill, const string &log);
	void PrintRemaining();
	struct MatchResult;
	void PrintMatchResult(const MatchResult &result);
	void ReadDate(const string &line);
	void ReadHour(const string &line);
	void ReadUid(const string &line);
	void ReadMin(const string &line);
	void ReadField(istream &in, void (Loader::*f)(const string &line));
	void Record();
	void Clear();
	string Day, Hour, Uid, Min;
	enum Sequence {GET_DATE, GET_HOUR, GET_UID, GET_MIN, COMPLETE} Seq;
	// values read from parsing the statement
	struct Statement
	{
		Statement(time_t t, const string &uid, int m) :
			start_time(t), remote_uid(uid), min(m) {}
		bool operator<(const Statement &rhs) const
		{ return start_time < rhs.start_time; }
		time_t start_time;
		string remote_uid;
		int min;
	};
	// order by start_time
	typedef set<Statement> StatementType;
	StatementType BillSet;
	// Database record
	struct DBEntry
	{
		DBEntry() : id(-1), start_time(-1), end_time(-1) {}
		DBEntry(int record_id, time_t start, time_t end,
			const string &uid) :
			id(record_id), start_time(start), end_time(end),
			remote_uid(uid) {}
		bool operator<(const DBEntry &rhs) const
		{ return start_time < rhs.start_time; }
		int id;
		time_t start_time, end_time;
		string remote_uid;
	};
	typedef set<DBEntry> DBEntryType;
	DBEntryType DBEntrySet;
	struct MatchResult
	{
		MatchResult(const DBEntry &entry, const Statement &bill) :
			Entry(entry), Bill(bill) {}
		struct DBEntry Entry;
		struct Statement Bill;
		bool operator<(const MatchResult &rls) const
		{
			return Entry < rls.Entry;
		}
	};
	typedef set<MatchResult> MatchType;
	MatchType MatchSet;
	sqlite3 *Db;
};

Loader::~Loader()
{
	sqlite3_close(Db);
}

bool Loader::OpenDB(const char *file)
{
	if(sqlite3_open(file, &Db))
	{
		cerr << "Error opening " << file << ' ' << sqlite3_errmsg(Db)
			<< endl;
		return false;
	}
	return true;
}

bool Loader::Parse(const char *file)
{
	ifstream in(file);
	if(!in)
		return false;
	string line;
	Clear();
	bool comment=false;
	cout << "reading file " << file << endl;
	while(!in.eof() && !in.bad())
	{
		if(Seq==COMPLETE)
			Record();
		getline(in, line);
		if(comment)
		{
			if(line.find("-->") != line.npos)
				comment=false;
		}
		else if(line.find("<!--") != line.npos)
			comment=true;
		else if(line.find("la_startOnDate") != line.npos)
			ReadField(in, &Loader::ReadDate);
		else if(line.find("la_startOnHour_print") != line.npos)
			ReadField(in, &Loader::ReadHour);
		else if(line.find("la_billPresentation1") != line.npos)
			ReadField(in, &Loader::ReadUid);
		else if(line.find("la_minutes") != line.npos)
			ReadField(in, &Loader::ReadMin);
	}
	return true;
}

bool Loader::LoadDB()
{
	if(BillSet.empty())
	{
		cout << "Error: Bill set empty\n";
		return false;
	}
	sqlite3_stmt *stmt=NULL;
	if(sqlite3_prepare_v2(Db, "select Events.id, start_time, end_time, "
		"remote_uid "
		"from Events, Plugins "
		"where service_id = Plugins.id and Plugins.name='CALL' and "
		"local_uid='ring/tel/ring' and "
		"start_time >= ? and start_time <= ?", -1, &stmt, NULL))
	{
		cerr << "Error preparing statement " << sqlite3_errmsg(Db)
			<< endl;
		return false;
	}
	int margin=12*60*60;
	time_t start=BillSet.begin()->start_time - margin,
		stop=BillSet.rbegin()->start_time + margin;
	cout << "Looking at " << start << " to " << stop << endl;
	sqlite3_bind_int(stmt, 1, start);
	sqlite3_bind_int(stmt, 2, stop);
	DBEntry entry;
	const char *str;
	while(sqlite3_step(stmt) == SQLITE_ROW)
	{
		if(sqlite3_column_type(stmt, 0) != SQLITE_INTEGER ||
			sqlite3_column_type(stmt, 1) != SQLITE_INTEGER ||
			sqlite3_column_type(stmt, 2) != SQLITE_INTEGER ||
			sqlite3_column_type(stmt, 3) != SQLITE_TEXT)
		{
			cerr << "Unexpected table schema\n";
			return false;
		}
		entry.id=sqlite3_column_int(stmt, 0);
		entry.start_time=sqlite3_column_int(stmt, 1);
		entry.end_time=sqlite3_column_int(stmt, 2);
		str=(const char *)sqlite3_column_text(stmt, 3);
		if(str)
			entry.remote_uid=str;
		else
			continue;
		DBEntrySet.insert(entry);
	}
	sqlite3_finalize(stmt);
	cout << "Loaded " << DBEntrySet.size() << " rows from the database\n";
	return true;
}

void Loader::PrintMatches()
{
	cout << "Matching entries\n";
	for(MatchType::iterator iter=MatchSet.begin(); iter!=MatchSet.end();
		++iter)
	{
		PrintMatchResult(*iter);
	}
}

void Loader::CompareDuration()
{
	cout << "Matching entries with duration\n";
	for(MatchType::iterator iter=MatchSet.begin(); iter!=MatchSet.end();
		++iter)
	{
		if(iter->Entry.end_time)
			PrintMatchResult(*iter);
	}
}

bool Loader::Update()
{
	cout << "Updating database entries\n";
	sqlite3_stmt *stmt=NULL;
	if(sqlite3_prepare_v2(Db, "update Events set end_time = ? "
		"where id = ? and start_time = ? and end_time = '0';",
		-1, &stmt, NULL))
	{
		cerr << "Error preparing statement " << sqlite3_errmsg(Db)
			<< endl;
		return false;
	}
	sqlite3_exec(Db, "begin;", NULL, NULL, NULL);
	bool result=true;
	for(MatchType::iterator iter=MatchSet.begin(); iter!=MatchSet.end();
		++iter)
	{
		// only update entries without an end_time
		if(iter->Entry.end_time)
			continue;

		PrintMatchResult(*iter);
		const DBEntry &entry=iter->Entry;
		sqlite3_reset(stmt);
		sqlite3_bind_int(stmt, 1, iter->Bill.min*60 + entry.start_time);
		sqlite3_bind_int(stmt, 2, entry.id);
		sqlite3_bind_int(stmt, 3, entry.start_time);
		if(sqlite3_step(stmt) != SQLITE_DONE)
		{
			cerr << "Error executing statement "
				<< sqlite3_errmsg(Db) << endl;
			result=false;
			break;
		}
		if(sqlite3_changes(Db) != 1)
		{
			cerr << "Error " << sqlite3_changes(Db) << " rows "
				"were modified, but expected just 1\n";
			result=false;
			break;
		}
	}
	sqlite3_finalize(stmt);
	sqlite3_exec(Db, result ? "commit;" : "rollback;", NULL, NULL, NULL);
	return result;
}

void Loader::PrintMatchResult(const Loader::MatchResult &result)
{
	char buf[80]={0};
	const char *time_out="%m-%d-%Y %l:%M %p";
	struct tm *t;
	t=localtime(&result.Bill.start_time);
	strftime(buf, sizeof(buf)-1, time_out, t);
	cout << "bill " << buf << ' ' << result.Bill.remote_uid
		<< '\t' << result.Bill.min << endl;

	t=localtime(&result.Entry.start_time);
	strftime(buf, sizeof(buf)-1, time_out, t);
	int sec=result.Entry.end_time - result.Entry.start_time;
	if(sec<0)
		sec=0;
	cout << "log  " << buf << ' ' << result.Entry.remote_uid
		<< '\t' << (sec+59)/60 << endl;
}

bool Loader::Match()
{
	int sec[] = {60, 2*60, 5*60, 1*60*60 + 2*60, 2*60*60 + 2*60,
		3*60*60 + 2*60};
	for(size_t i=0; i<sizeof(sec)/sizeof(*sec); ++i)
	{
		Match(sec[i]);
		cout << "leniency " << sec[i] << " match " << MatchSet.size()
			<< endl;
	}
	cout << "Final counts statement " << BillSet.size()
		<< " log " << DBEntrySet.size() << " match " << MatchSet.size()
		<< endl;
	PrintRemaining();
	return !MatchSet.empty();
}

bool Loader::Match(int sec)
{
	StatementType::iterator siter=BillSet.begin();
	DBEntryType::iterator liter=DBEntrySet.begin();
	while(siter!=BillSet.end() && liter!=DBEntrySet.end())
	{
		if(abs(siter->start_time - liter->start_time) < sec &&
			UidMatch(siter->remote_uid, liter->remote_uid))
		{
			MatchSet.insert(MatchResult(*liter, *siter));
			BillSet.erase(*siter++);
			DBEntrySet.erase(*liter++);
			continue;
		}
		if(siter->start_time < liter->start_time)
			++siter;
		else
			++liter;
	}
	return false;
}

void Loader::PrintRemaining()
{
	char buf[80]={0};
	const char *time_out="%m-%d-%Y %l:%M %p";
	struct tm *t;
	StatementType::iterator siter=BillSet.begin();
	DBEntryType::iterator liter=DBEntrySet.begin();
	cout << "Unmatched entries\n";
	while(siter!=BillSet.end() && liter!=DBEntrySet.end())
	{
		if(siter->start_time < liter->start_time)
		{
			t=localtime(&siter->start_time);
			strftime(buf, sizeof(buf)-1, time_out, t);
			cout << "bill " << buf << ' ' << siter->remote_uid
				<< '\t' << siter->min << endl;
			++siter;
		}
		else
		{
			t=localtime(&liter->start_time);
			strftime(buf, sizeof(buf)-1, time_out, t);
			int sec=liter->end_time - liter->start_time;
			if(sec<0)
				sec=0;
			cout << "log  " << buf << ' ' << liter->remote_uid
				<< '\t' << (sec+59)/60 << endl;
			++liter;
		}
	}
}

bool Loader::Digits(const std::string &name)
{
	for(size_t i=0; i<name.size(); ++i)
		if(!isdigit(name[i]))
			return false;
	return true;
}

string Loader::RemoveDash(const std::string &in)
{
	if(in.find('-')==in.npos)
		return in;
	string out;
	for(size_t i=0; i<in.size(); ++i)
		if(in[i]!='-')
			out+=in[i];
	return out;
}

string Loader::Normalize(string in)
{
	in=RemoveDash(in);
	if(in.size()==12 && in[0]=='+' && in[1]=='1')
	{
		string sub=in.substr(2);
		if(Digits(sub))
			return sub;
		return in;
	}
	else if(in.size()==11 && in[0]=='1' && Digits(in))
		return in.substr(1);
	return in;
}

bool Loader::UidMatch(const string &bill, const string &log)
{
	if(bill==log)
		return true;
	// For T-Mobile 123 shows up on the bill while the log lists
	// the full number.
	if(bill=="123" && (log=="+18056377243" || log=="+18056377249"))
		return true;
	string bstr=Normalize(bill);
	string lstr=Normalize(log);
	if(bstr==lstr)
		return true;
	return false;
}

#define CHECK_STATE(STATE) \
if(Seq!=STATE) \
	{ cerr << "bad state " << Seq << " expected " << STATE << " data "\
		"'" << Day << "' '" << Hour << "' '" << Uid \
		<< "' '" << Min << "' line '" << line << "'\n"; return; } \
else \
	{ Seq=(Sequence)(STATE+1); }

void Loader::ReadDate(const string &line)
{
	CHECK_STATE(GET_DATE)
	Day=line;
}

void Loader::ReadHour(const string &line)
{
	CHECK_STATE(GET_HOUR)
	Hour=line;
}

void Loader::ReadUid(const string &line)
{
	CHECK_STATE(GET_UID)
	Uid=line;
}

void Loader::ReadMin(const string &line)
{
	CHECK_STATE(GET_MIN)
	Min=line;
}

// filter out headers and other fields that follow that line
void Loader::ReadField(istream &in, void (Loader::*f)(const string &line))
{
	string line;
	getline(in, line);
	if(line.find("<b>") != line.npos || line.find("<img") != line.npos ||
		line.find("<span") != line.npos)
		return;
	size_t start=line.find_first_not_of(" \t\r");
	size_t end=line.find_last_not_of(" \t\r");
	if(start==line.npos || end==line.npos || end<start)
		return;
	(this->*f)(line.substr(start, end-start+1));
}

void Loader::Record()
{
	if(Day.empty() || Hour.empty() || Uid.empty()
		|| Min.empty())
	{
		cerr << "Incomplete record '" << Day << "' '" << Hour
			<< "' '" << Uid << "' '" << Min << "'\n";
		Clear();
		return;
	}
	cout << (Day[1]=='/' ? " " : "")  << Day << ' '
		<< (Hour[1]==':' ? " " : "") << Hour << ' ' << Uid << ' '
		<< Min << endl;
	string date=Day + " " + Hour;
	struct tm t={0};
	// tm isn't just an output data structure, tell it to use DST
	// according to the current timezone rules at that time.
	t.tm_isdst=-1;
	const char *format="%m/%d/%y %I:%M %p";
	const char *result=strptime(date.c_str(), format, &t);
	if(!result)
	{
		cerr << "Failed to parse date\n";
		Clear();
		return;
	}
	// statement
	// The date and time for all messaging and data corresponds to
	// Pacific Time (PST/PDT).
	// except they mostly are local time
	/*
	t.tm_hour+=2;
	t.tm_isdst=-1;
	*/
	time_t tsec=mktime(&t);

	string remote_uid=RemoveDash(Uid);
	int min=atoi(Min.c_str());

	char buf[80]={0};
	//const char *out="%m-%d-%Y %H:%M";
	const char *out="%m-%d-%y %l:%M %p";
	strftime(buf, sizeof(buf)-1, out, &t);
	cout << buf << " ( " << tsec << " ) " << remote_uid << ' ' << min
		<< endl;
	BillSet.insert(Statement(tsec, remote_uid, min));
	Clear();
}

void Loader::Clear()
{
	Day.clear();
	Hour.clear();
	Uid.clear();
	Min.clear();
	Seq=GET_DATE;
}

// This program is designed to Read the html statement page to get the
// start_time, remote_uid, and minutes to load the missing duration into the
// N900 call log. 
int main(int argc, char **argv)
{
	Loader loader;
	if(argc>=2)
	{
		if(!loader.OpenDB(argv[1]))
			return 1;
	}
	bool update=false;
	for(int i=2; i<argc; ++i)
	{
		if(!strcmp(argv[i], "--update"))
			update=true;
		else if(!loader.Parse(argv[i]))
			return 1;
	}
	if(!loader.LoadDB())
		return 1;
	if(!loader.Match())
		return 1;
	loader.PrintMatches();
	loader.CompareDuration();
	if(update && !loader.Update())
		return 1;
	return 0;
}
