/* MinuteMonitor, log duration and display usage for the N900
 * Copyright (c) 2012 David Fries <David@Fries.net> http://david.fries.net/
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// QtQuick doesn't ship by default on N900 (and isn't required here)
//import QtQuick 1.0
import Qt 4.7
import MinuteMonitor 0.1

// Display color coded (by call type) log entries 
Rectangle
{
	id: labelList
	anchors.fill: parent
	color: "#353535"
	// reminder, store is not available when constructed, don't bind
	// properties until onCompleted
	Component.onCompleted : 
	{
		// The Connections didn't seem to work, but this callback
		// does and gets called from the correct object.
		windowStack.hidden.connect(function()
		{
			labelList.destroy()
		})
	}

	/* direction to select the image, pass 0 to n increment until the image
	 * is found or null is returned
	 * returns a url for the requested direction
	 */
	function get_imagename(direction, pass)
	{
		var file=null
		var base=null
		switch(direction)
		{
		// used for the contact list where there isn't an in/out
		case CallType.CALL_NEITHER:
			file=null
			break
		case CallType.CALL_IN:
			file=image_in
			base="general_received.png"
			break
		case CallType.CALL_OUT:
			file=image_out
			base="general_sent.png"
			break
		case CallType.CALL_BOTH:
			file=image_both
			base="addressbook_recent_contacts_group.png"
			break
		default:
			file=image_unknown
			base="filemanager_unknown_file.png"
		}
		if(!file || file=="")
			return file

		// try a limited number of location rules
		var prefix="file://"
		// location to scratchbox when not in scratchbox
		var scratchbox=
		"/scratchbox/users/david_armel/targets/FREMANTLE_ARMEL"
		var maemo="/usr/share/icons/hicolor/32x32/hildon/"

		// first attempt is the expected maemo location
		if(file==do_search)
		{
			file=prefix + maemo + base
			set_imagename(direction, file)
			return file
		}

		switch(pass)
		{
		case 0:
			// cached value
			break
		case 1:
			file=prefix + scratchbox + maemo + base
			break
		default:
			file=null
		}
		return file
	}
	// cache the image name, call when pass > 0
	function set_imagename(direction, file)
	{
		// avoid "Unable to assign null to QString"
		if(!file)
			file=""
		switch(direction)
		{
		// used for the contact list where there isn't an in/out
		case CallType.CALL_NEITHER:
			break
		case CallType.CALL_IN:
			image_in=file
			break
		case CallType.CALL_OUT:
			image_out=file
			break
		case CallType.CALL_BOTH:
			image_both=file
			break
		default:
			image_unknown=file
		}
	}

	function call_direction(direction)
	{
		switch(direction)
		{
		// also used for the contact list where there isn't an in/out
		case CallType.CALL_NEITHER:
			return ""
		case CallType.CALL_IN:
			return "in"
		case CallType.CALL_OUT:
			return "out"
		case CallType.CALL_BOTH:
			return "in/out"
		default:
			return "?"
		}
	}
	property variant image_loader: null
	property variant text_loader: null
	property real direction_width: 0
	function createDirectionLabel(parent, button, model)
	{
		if(model.direction == CallType.CALL_NEITHER)
			return
		var image=null
		for(var pass=0; ; ++pass)
		{
			var file = get_imagename(model.direction, pass)
			if(!file)
				break
			if(!image)
			{
				if(image_loader==null)
				{
					image_loader=Qt.createComponent(
						"LabelImage.qml")
					if(image_loader.status!=Component.Ready)
					{
						print(image_loader.
							errorString())
						image_loader=""
					}
				}
				if(image_loader == "")
					break
				image = image_loader.createObject(parent)
				// if this fails don't bother trying to fall
				// back to Text
				if(!image)
					return
			}
			image.source = file
			if(image.status == Image.Ready)
			{
				if(direction_width < image.width)
					direction_width = image.width
				image.anchors.verticalCenter =
					parent.verticalCenter
				// cache to avoid the search next time
				if(pass>0)
					set_imagename(model.direction, file)
				return
			}
		}
		if(image)
		{
			image.destroy()
			set_imagename(model.direction, null)
		}
		if(text_loader==null)
		{
			text_loader=Qt.createComponent("LabelText.qml")
			if(text_loader.status!=Component.Ready)
			{
				print(text_loader.errorString())
				text_loader=""
			}
		}
		if(text_loader == "")
			return
		var text=text_loader.createObject(parent)
		if(!text)
			return
		text.text = call_direction(model.direction)
		text.color = "white"
		if(direction_width < text.width)
			direction_width = text.width
		text.anchors.verticalCenter = parent.verticalCenter
	}

	ListView
	{
		id: listView
		model: labelModel
		anchors.fill: parent
		pressDelay: press_delay
		focus: true
		keyNavigationWraps: true // quick wrap around
		highlight: Rectangle { color: selected_color }
		highlightMoveDuration: 1
		delegate:
		Item
		{
			id: button
			width: parent.width
			height: Math.max(name.height+uid.height,
				start_time.height+minutes.height) + 6
			// Give each row a black background.
			Rectangle
			{
				// set to draw before the highlight component
				parent: button.parent
				z: -100
				y: button.y
				height: button.height
				width: button.width
				color: "black"
			}
			Item
			{
				id: outgoing
				anchors.verticalCenter: button.verticalCenter
				// select between an image, text, or nothing
				Component.onCompleted : createDirectionLabel(
					outgoing, button, model)
			}
			Text
			{
				id: name
				y: 3
				x: direction_width + 5
				text: model.remote_name
				color: "white"
			}
			Text
			{
				id: uid
				y: 3 + name.height
				x: name.x + 10
				text: model.remote_uid
				color: "white"
				font.pointSize: name.font.pointSize*.8
			}
			Rectangle
			{
				color:
					model.type == CallType.CALL_COST ?
						"red" :
					model.type == CallType.CALL_FREE ?
						"yellow" :
					model.type == CallType.CALL_VOIP ?
						"green" : "black"
				anchors.fill: call_type
				visible: model.type <= CallType.CALL_VOIP
			}
			Text
			{
				id: call_type
				x: parent.width - width - 25
				y: 3
				text: model.type == CallType.CALL_COST ?
						" Cost " :
					model.type == CallType.CALL_FREE ?
						" Free " :
					model.type == CallType.CALL_VOIP ?
						" VoIP " : " unknown "
				color: "black"
			}
			Text
			{
				id: minutes
				x: call_type.x - width - 25
				y: 3
				text: model.minutes
				color: "white"
			}
			Rectangle
			{
				color: 
					(model.rate & CallType.DAYTIME_RATE ?
						"Red" :
					model.rate & CallType.EVENING_RATE ?
						"Yellow" : "black")
				anchors.fill: call_rate
				visible: call_rate.visible
			}
			Text
			{
				id: call_rate
				anchors.top: start_time.top
				anchors.right: start_time.left
				anchors.rightMargin: 5
				text: (model.rate & CallType.DAYTIME_RATE ?
						" Day " : "") +
					(model.rate & CallType.EVENING_RATE ?
						" Evening " : "")
				font.pointSize: name.font.pointSize*.8
				visible: text != ""
			}
			Text
			{
				id: start_time
				x: parent.width - width - 40
				y: 3 + minutes.height
				text: model.start_time
				font.pointSize: name.font.pointSize*.8
				color: "white"
			}
			Rectangle
			{
				height: 1
				x : 20
				width: parent.width - 2*x
				color: divider_color
			}
			MouseArea
			{
				anchors.fill: button
				onPressed: listView.currentIndex=index
				/*
				onReleased: listView.currentIndex=-1
				onCanceled: listView.currentIndex=-1
				*/
				function edit()
				{
					create("EditType.qml", function(obj)
					{
						obj.parent=labelList
						obj.model=model
						obj.update()
					})
				}
				onClicked: edit()
				focus: listView.currentIndex==index
				Keys.onEnterPressed: edit()
			}
		}
	}
	ScrollDecorator
	{
		flickableItem: listView
	}
}
